#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <netinet/in.h>

#include <zlib.h>
#include <unistd.h>
#include <sys/stat.h>

#include <stdarg.h>
#include <errno.h>

/* Compile:  gcc -o skel -g skel.c -lz
   Debug:    gdb ./skel                 */

/* Nevil's declarations for the Dag trace records ... */

/* Sizes for PC Linux:      int=4, long=4, long long=8
   Sizes for Digital Unix:  int=4, long=8, long long=8 */

#define SIZEOF_LONG 4  /* PC */

typedef unsigned char  Bit8;
typedef unsigned short Bit16;
#if SIZEOF_LONG == 4
typedef unsigned long  Bit32;
#elif SIZEOF_LONG == 8
typedef unsigned int  Bit32;
#endif

struct ether_hdr {  
   unsigned char dest[6];    /* MAC addresses */
   unsigned char source[6];
   Bit16 type;               /* Length or protocol type */
   };

/* Dag trace file record layouts ...
      Note that they all use 64-byte fixed-length records */

struct snap_hdr {
   Bit16 lsap;           /* AAAA */
   Bit8 ctrl;            /* 03 */
   Bit8 oid1,oid2,oid3;  /* 00 00 00 */
   Bit16 etype;          /* 0800 */
   };

struct dag_atm {  /* Dag record for ATM (first cell only, a la OC3MON) */
   Bit32 crc;           /* Always zero */
   Bit32 atm_hdr;       /* VPI:VCI, etc */
   struct snap_hdr sh;  /* LLC snap header (8 bytes) */
   Bit8 pload[40];      /* First 40 bytes of IP packet */
   };

struct dag_ether {  /* Dag record for Ethernet */
   Bit16 wlen;          /* Wire length (bytes) */
   struct ether_hdr eh; /* Ethernet MAC header */
   Bit8 pload[40];
   };

struct dag_pos {  /* Dag record for PoS */
   Bit32 slen;          /* Snap length, must be 64 for now */
   Bit32 wlen;          /* Bytes on the wire, seems to include FCS */
   Bit8 chdlc;          /* Cisco HDLC header */
   Bit8 unused;         /* Padding */
   Bit16 etype;         /* EtherType */
   Bit8 pload[44];      /* 4 more bytes than ATM and Ether */
   };

struct dag_ext_pos {  /* Extensible Type 1, PoS HDLC */
   Bit8 chdlc;          /* Cisco HDLC header (this + next 3 bytes) */
   Bit8 unused;         /* Padding */
   Bit16 etype;         /* EtherType */
   Bit8 pload[44];      /* rlen-20, rlen default is 64 */
   };

struct dag_ext {      /* 'Extensible' Dag record */
   Bit8 type;           /* Dag type (TYPE_ defines, in dag.h) */
   Bit8 flags;          /* Dag flags (DF_ defines, below) */
   Bit16 rlen;          /* Dag data length */
   Bit16 lctr;          /* Loss counter, sticks at 0xffff */
   Bit16 wlen;          /* Bytes on the wire */
   union {
      struct dag_ext_pos pos;
      } e;
   };

#define DF_INTERFACE  0x03  /* Capture interface 0-3 */
#define DF_VBL_LEN    0x04  /* Variable length records */
#define DF_TRUNCATED  0x08  /* Record truncated (insufficient buffer space) */
#define DF_RX_ERROR   0x10  /* Rx error (link layer) */
#define DF_DS_ERROR   0x20  /* Internal error */
#define DF_RSRV       0xC0  /* Reserved */

#define TT_ATM    0  /* Dag trace file record types */
#define TT_ETHER  1
#define TT_POS    2
#define TT_EXT    3

#define DAG_DEFAULT_SLEN   48  /* These match the Legacy records */
#define DAG_DEFAULT_VARLEN  0

struct dag_rec {
   long long ts;        /* Dag timestamp (64-bit) */
   union {
      struct dag_atm atm;
      struct dag_ether eth;
      struct dag_pos pos;
      struct dag_ext ext;
      } d;
   };

/* Conversion functions: s and us from Dag (64-bit) timestamp */
#define DAGT_S(dts)   (long)((dts) >> 32)
#define DAGT_US(dts)  (long)((((dts) & 0xffffffffLL) * 1000000) >> 32)

#define US_DAGT(us)   (((long long)(us) << 32)/1000000LL)

#define PT_ICMP       1  /* IP protocol types */
#define PT_TCP        6
#define PT_UDP       17

void quack(int where) { }

static char     *prog;
static void     set_argv0(char *argv0);
void     	panic(char *fmt, ...) __attribute__((noreturn, format (printf, 1, 2)));
static void     usage(void)	      __attribute__((noreturn));

static char	*mtempl = "wb%u";
static char	mode[16];

int verbose, dag_type, level, trace_limit;
static char *raw_fn, *out_fn;
static char *out_suffix = "skel.gz";

static void process(void);

int main(int argc, char *argv[])
{
   int opt, i, ofn_len;

   verbose = trace_limit = dag_type = TT_EXT;  /* Defaults */
   out_fn = raw_fn = NULL;
   level = 1;

   set_argv0(argv[0]);
   opterr = 0;
   while((opt = getopt(argc, argv, ":hv123456789g:o:t:")) != EOF)
      switch(opt) {
      case 'g':  /* -g dag type, default = 0 (atm) */
         dag_type = strtoul(optarg, NULL, 0);
         if (dag_type < TT_ATM || dag_type > TT_POS)
            panic("Dag type (-g) not 0 (ATM), 1 (Ether) or 2 (PoS)\n");
         break;
      case 'h':
         usage();
      case 'o':
         out_fn = optarg;
         break;
      case 't':  /* -t n to trace n packets */
         trace_limit = strtoul(optarg, NULL, 0);
         break;
      case '1': case '2': case '3': case '4': case '5':
      case '6': case '7': case '8': case '9':
         level = opt - '0';	/* poor man's strtoul(3) */
         break;
      case 'v':
         ++verbose;
         break;
      case '?':
         panic("unrecognized option -%c, see -h for usage\n", optopt);
      default:
         panic("missing argument to option -%c, see -h for usage\n", optopt);
      }
   argc -= optind;  argv += optind;
   if (argc > 0) {
      raw_fn = strdup(argv[0]);
      ofn_len = strlen(raw_fn)+1+strlen(out_suffix)+1;
      out_fn = (char *)malloc(ofn_len);
      (void)snprintf(out_fn,ofn_len, "%s.%s", raw_fn,out_suffix);
      }

   if (out_fn == NULL) panic("No output file specified (-o fn)");
   snprintf(mode, sizeof(mode), mtempl, level);

   if (verbose != 0) {
      fprintf(stderr, 
         "argc=%d, dag_type=%d, level=%d, raw_file=%s, output mode=%s\n",
         argc, dag_type, level, argc > 0 ? raw_fn : "stdin", mode);
      fprintf(stderr, "  Anonymised output written to %s\n", out_fn);
      }

   process();
   }


void process(void)
{     
   gzFile gfp, ofp;
   int r, first, errnum;
   long long ts;

#define RAW_BLK_SZ  4096
   Bit8 dbuf[RAW_BLK_SZ];
   int n;
   struct dag_rec *drp;
   int rl, nib, left_ib;

#if 0
   Bit8 *hdrp;
   int prot, dx;
   Bit32 saddr,daddr;
   Bit16 sport, dport;
#endif

   if (raw_fn != NULL) {
      if ((gfp = gzopen(raw_fn,"r")) == NULL)
         panic("Can't open raw dag file %s\n", raw_fn);
      }
   else if ((gfp = gzdopen(STDIN_FILENO,"r")) == NULL)
      panic("Can't open stdin!\n");

#define WRITING 0
#if WRITING
   if((ofp = gzopen(out_fn, mode)) == NULL)
      panic("Can't gzopen output file %s %s: %s and %s\n",
         out_fn, mode, strerror(errno), gzerror(ofp, &errnum));
#endif

   first = 1;  left_ib = 0;
   for (;;) {
      if ((r = gzread(gfp, &dbuf[left_ib], sizeof(dbuf)-left_ib)) == 0) 
         break;  /* End of file */
      r += left_ib;

      if (first) {
	 ts = *(long long *)dbuf;
         fprintf(stderr, "First record at %10d.%06d\n",
	    DAGT_S(ts), DAGT_US(ts));
	 first = 0;
	 }

      quack(1);
//      printf("--- r=%d\n\n", r);
      nib = 0;  drp = (struct dag_rec *)dbuf;
      for (;;) {
         rl = ntohs(drp->d.ext.rlen);

         if (nib + rl > r) {  /* Part of a record */
 	    left_ib = r - nib;
            memcpy(dbuf, &dbuf[nib], left_ib);
            break;
            } 
	 else {
#if 0
            printf("%8d %10d.%06d: %02x-%02x %u %u %u\n",
               n, DAGT_S(drp->ts), DAGT_US(drp->ts),
	       drp->d.ext.type, drp->d.ext.flags,
	       rl, ntohs(drp->d.ext.lctr), ntohs(drp->d.ext.wlen) );
#endif
#if WRITING
            if (gzwrite(ofp, drp, rl) != rl)
               panic("gzopen write error %s: %s and %s\n",
                  out_fn, strerror(errno), gzerror(ofp, &errnum));
#endif
            nib += rl;
            drp = (struct dag_rec *)( ((Bit8 *)drp) + rl );
            ++n;
            }
         }
//      printf("\n");
//      if (n >= 10) break;
      }

   gzclose(gfp);
#if WRITING
   gzclose(ofp);
#endif
   ts = *(long long *)drp;
   fprintf(stderr, "Last record at %10d.%06d\n",
      DAGT_S(ts), DAGT_US(ts));
   fprintf(stderr, ">>> %u records\n", n);

   if (r != 0) fprintf(stderr, "gzread failed, returned %d\n", r);

#if 0
   n = 0;
   while ((r = fread(&dbuf, sizeof(Bit8), RAW_BLK_SZ, raw)) != 0) {
     quack(1);

     for (j = 0; j != r; ++j) {
         drp = &dbuf[j];
         hdrp = (Bit8 *)&drp->d.eth.pload;  /* IP packet header */

         prot = hdrp[9];  /* IPv4 Protocol Type */
         saddr = ntohl(*(Bit32 *)&hdrp[12]);  /* IPv4 source */
         daddr = ntohl(*(Bit32 *)&hdrp[16]);  /* IPv4 dest */
            /* IP Addresses are anonymised on the dag trace */
         
         dx = (hdrp[0] & 0x0F) << 2;  /* HLEN */
         /* Data offset.  HLEN = IPv4 header length in 32-bit units */
         if (prot == PT_ICMP) {
            sport = ntohs(hdrp[dx]);  /* Type field */
            dport = ntohs(hdrp[dx+1]);  /* Code field */
            }
         else if (prot == PT_TCP || prot == PT_UDP) {
            sport = ntohs(*(Bit16 *)&hdrp[dx]);  /* Source port */
            dport = ntohs(*(Bit16 *)&hdrp[dx+2]);  /* Dest port */
            }
         else sport = dport = 0;

         printf("%8d %10d.%06d: %3d, %08x->%08x, %5d->%5d\n",
            n, DAGT_S(drp->ts), DAGT_US(drp->ts),
            prot, saddr,daddr, sport,dport);

         ++n;
         if (n == 5) exit(3);
         }
      }
#endif
   }

/* Joerg's helper functions */

static void
set_argv0(char *argv0)
{
	if((prog = strrchr(argv0, '/')) == NULL)
		prog = argv0;
	else
		prog++;
}

void
panic(char *fmt, ...)
{
	va_list ap;

	fprintf(stderr, "%s: panic: ", prog);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	exit(1);
}

static char usgtxt[] =
   "%s: test file scanner for dag data files\n"
   "usage: %s [-gop] input-file\n"
   "    -g m   m=0 ATM, m=1 Ether, m=2 POS (3 EXT default)\n"
   "    -o fn  specifies output filename (required)\n"  
   "  reads stdin if no filename given\n"
   ;

static void
usage()
{
	fprintf(stderr, usgtxt, prog, prog);
	exit(1);
}
